import logging
import re
import sys

import ipywidgets
from scrapydo.utils import highlight

from pygments.formatters import TerminalFormatter
from pygments.lexers import PythonTracebackLexer, PythonLexer
from fabric.colors import blue, cyan, green, magenta, red

lexer = PythonLexer()
python_traceback_lexer = PythonTracebackLexer()
terminal_formatter = TerminalFormatter()
re_color_codes = re.compile(r'\033\[(\d;)?\d+m')


LEVELS = {
    'INFO': blue('INFO'),
    'DEBUG': blue('DEBUG'),
    'WARNING': red('WARN', bold=True),
    'CRITICAL': magenta('CRIT'),
    'ERROR': red('ERROR'),
}


class ScrapyFormatter(logging.Formatter):
    def __init__(self):
        logging.Formatter.__init__(self)

    def format(self, record):
        s = '%(levelname)s %(name)s %(message)s' % {
            'levelname': LEVELS[record.levelname],
            'name': cyan(record.name),
            'message': record.getMessage()
        }

        if record.exc_info:
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            try:
                exc_text = record.exc_text
            except UnicodeError:
                exc_text = record.exc_text.decode(sys.getfilesystemencoding())
            s += highlight(exc_text, python_traceback_lexer, terminal_formatter)

        return s


class OutputWidgetHandler(logging.Handler):

    def __init__(self):
        super(OutputWidgetHandler, self).__init__()
        self.out = ipywidgets.Output()

    def emit(self, record):
        formatted_record = self.format(record)
        new_output = {
            'name': 'stdout',
            'output_type': 'stream',
            'text': formatted_record+'\n'
        }
        self.out.outputs = self.out.outputs + (new_output, )

    def clear_logs(self):
        self.out.clear_output()


def create_handler():
    handler = OutputWidgetHandler()
    handler.setFormatter(ScrapyFormatter())
    logger = logging.getLogger('scrapy')
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    return handler


def highlight_html(code):
    return highlight(code, output_wrapper=str).replace('<body>\n<h2></h2>\n\n', '<body>')


import scrapydo
from crawlab.proxy import ZapProxy
from crawlab.utils import create_handler


class Crawler:
    def __init__(self, ):
        self.proxy = ZapProxy()
        self.log_handlers = {}
        scrapydo.setup()

    def get_proxy(self):
        return self.proxy.get_url()

    def get_handler(self, key):
        return self.log_handlers.get(key, create_handler())

    def get_messages(self, url):
        return self.proxy.get_messages(url)

    def configure_proxy(self, request):
        if not self.proxy.started:
            started = self.proxy.start_server()
            if not started:
                raise Exception('Proxy not started')
        self.proxy.delete_messages(request.url)

    def fetch(self, request):
        if request.meta.get('proxy') == self.get_proxy():
            self.configure_proxy(request)

        callback = None
        if request.callback:
            callback = request.callback
            request.callback = None

        if not self.log_handlers.get(request.url):
            self.log_handlers[request.url] = create_handler()
        self.log_handlers[request.url].clear_logs()

        response = scrapydo.fetch(request)

        if callback:
            callback(response)

        return response



import html

from IPython.display import display
import ipywidgets
from ipywidgets import Layout, HTML
from traitlets import Unicode, Dict
from bs4 import BeautifulSoup
import base64

from crawlab.utils import highlight_html


@ipywidgets.register
class NetworkWidget(ipywidgets.VBox):
    
    def __init__(self, messages):
        messages_widgets = []
        for message in messages:
            messages_widgets += [self._network_box(message['request']), self._network_box(message['response'])]
        super(NetworkWidget, self).__init__(messages_widgets)

    @staticmethod
    def _box(box_widgets):
        return ipywidgets.VBox(box_widgets, layout=Layout(
            border='1px solid gray', padding='3px', margin='3px'))

    def _network_box(self, message):
        return self._box([
            self._title(message['title']),
            self._title_text_box('Header', message['header']),
            self._title_text_box('Body', message['body']),
        ])

    @staticmethod
    def _title(title):
        return ipywidgets.HTML(value='<b>' + title + '</b>')

    def _title_text_box(self, title, text):
        text_box_widgets = [self._title(title)]
        if text:
            text_box_widgets.append(self._box([ipywidgets.HTML(value=text)]))
        return self._box(text_box_widgets)


@ipywidgets.register
class BrowserWidget(ipywidgets.DOMWidget):
    _view_name = Unicode('BrowserView').tag(sync=True)
    _view_module = Unicode('crawlab').tag(sync=True)
    
    srcdoc = Unicode('srcdoc').tag(sync=True)
    
    def __init__(self, srcdoc='', **kwargs):
        self.srcdoc = srcdoc
        
        super(BrowserWidget, self).__init__(**kwargs)


@ipywidgets.register
class HTMLResponseWidget(ipywidgets.Tab):
    def __init__(self, response, messages, log_handler):
        children = [
            ('Browser', BrowserWidget(srcdoc=html.escape(response.text))),
            ('Code', ipywidgets.HTML(highlight_html(response.text))),
            ('Logs', log_handler.out),
            ('Network', NetworkWidget(messages=messages))
        ]

        super(HTMLResponseWidget, self).__init__([widget for name, widget in children], layout={'height': '300px'})

        for i, widget in enumerate(children):
            self.set_title(i, widget[0])


@ipywidgets.register
class ResponseWidget(ipywidgets.Tab):
    def __init__(self, response, messages, log_handler):
        content_type = response.headers['Content-Type'].decode().split(';')[0]
        url = 'data:' + content_type + ';base64,' + base64.b64encode(response.body).decode()
        html_str = '<iframe src="{url}" width="100%", height="280px"></iframe>'.format(url=url)
        children = [
            ('Browser', ipywidgets.HTML(html_str)),
            ('Logs', log_handler.out),
            ('Network', NetworkWidget(messages=messages))
        ]

        super(ResponseWidget, self).__init__([widget for name, widget in children], layout={'height': '300px'})

        for i, widget in enumerate(children):
            self.set_title(i, widget[0])


@ipywidgets.register
class SelectorResultWidget(ipywidgets.VBox):
    def __init__(self, selector):
        self.title = 'Result'
        results = []
        for data_str in selector.extract():
            code_widget = ipywidgets.HTML(highlight_html(data_str))
            box = ipywidgets.VBox([code_widget], layout=Layout(border='1px solid gray', padding='3px', margin='3px'))
            results.append(box)
        super(SelectorResultWidget, self).__init__(results)


@ipywidgets.register
class SelectorWidget(ipywidgets.Tab):

    def __init__(self, selector):
        children = [
            ('Result', SelectorResultWidget(selector)),
            ('Code', ipywidgets.HTML(self._format_code(selector)))
        ]

        super(SelectorWidget, self).__init__([widget for name, widget in children])

        for i, widget in enumerate(children):
            self.set_title(i, widget[0])

    def _format_code(self, selector):
        if len(selector) > 0:
            start_mark = '{mark}'
            end_mark = '{/mark}'

            marked_code = ''.join(selector[0].xpath('/*').extract())
            for tag in selector.extract():
                marked_code = marked_code.replace(tag, start_mark + tag + end_mark)

            highlighted_code = highlight_html(marked_code)
            prettify_html = BeautifulSoup(highlighted_code, 'html.parser').prettify()

            return prettify_html\
                .replace(start_mark, '<mark style="background-color: yellow;">').replace(end_mark, '</mark>')

        return 'Not found'


class JsonWidget(ipywidgets.DOMWidget):
    _view_name = Unicode('JsonView').tag(sync=True)
    _view_module = Unicode('crawlab').tag(sync=True)
    
    data = Dict({}).tag(sync=True)
    
    def __init__(self, data, **kwargs):
        self.data = dict(data)
        css = """
            .mark-water{
                color:#bbb;
            }
            .node-content-wrapper{
                font-family: 'Quicksand', sans-serif;
                background-color:#fff;
            }
            .node-content-wrapper ul{
                border-left:1px dotted #ccc;
                list-style:none;
                padding-left:25px;
                margin:0px;
            }
            .node-content-wrapper ul li{
                list-style:none;
                border-bottom:0;
                padding-bottom:0
            }
            .node-hgl-path{
                background-color:#fefbdf;
            }
            .node-bracket{
                font-weight:bold;
                display:inline-block;
                cursor:pointer;
            }
            .node-bracket:hover{
                color:#999;
            }
            .leaft-container{
                width:100%;
                max-width:300px;
                height:100%;
            }
            .title{ color:#ccc;}
            .string{ color:#080;}
            .number{ color:#ccaa00;}
            .boolean{ color:#1979d3;}
            .date{ color:#aa6655;}
            .null{ color:#ff5050;}
        """
        display(HTML('<style>' + css + '</style>'))

        super(JsonWidget, self).__init__(**kwargs)




import uuid


class ContextManager:
    def __init__(self, shell):
        self.ip = shell

    def get(self, key):
        if not self.ip.user_ns.get(key):
            raise Exception('{} not found'.format(key))
        return self.ip.user_ns[key]

    def set(self, key, value):
        self.ip.user_ns[key] = value

    def execute_and_get(self, command):
        tmp_key_name = '___tmp_' + str(uuid.uuid4().int)[:10]
        self.ip.kernel.do_execute('{}={}'.format(tmp_key_name, command), False, store_history=False)
        return self.get(tmp_key_name)


import validators
import re

from IPython.core.magic import line_magic, magics_class, Magics

import scrapy
from crawlab.context_manager import ContextManager


@magics_class
class MagicCrawler(Magics):
    def __init__(self, shell, crawler):
        super(MagicCrawler, self).__init__(shell)

        self.cm = ContextManager(shell)
        self.crawler = crawler

    @line_magic
    def fetch(self, request):
        if type(request) is str:
            request = self._get_request_by_cmd(request)

        response = self.crawler.fetch(request)
        self.cm.set('response', response)

        return response

    @line_magic
    def xpath(self, code):
        return self._execute_selector_fn('xpath', code)

    @line_magic
    def css(self, code):
        return self._execute_selector_fn('css', code)

    @line_magic
    def item(self, code=None, **kwargs):
        if code and type(code) is str:
            kwargs = self.cm.execute_and_get('dict({})'.format(code))

        item_attrs = {k: scrapy.Field() for k in kwargs.keys()}
        item_cl = type('MagicItem', (scrapy.Item,), item_attrs)
        return item_cl(**kwargs)

    def _get_request_by_cmd(self, request):
        if validators.url(request):
            return scrapy.Request(request)
        return self.cm.execute_and_get(request)

    def _execute_selector_fn(self, selector_fn, code):
        commands = self._extract_commands(code)
        result = getattr(self.cm.get('response'), selector_fn)(commands.get('expr'))
        target = commands.get('target')
        if target:
            self.cm.set(target, result)
        return result

    @staticmethod
    def _extract_commands(code):
        match = re.match('(?P<expr>[^\s\>]*)(\s(?P<sep>\>)(\s*)(?P<target>.*))?', code)
        if not match:
            raise Exception('Wrong command')
        return match.groupdict()



from crawlab.crawler import Crawler
from crawlab.magic_crawler import MagicCrawler

from scrapy.http.response.html import HtmlResponse
from scrapy.http.response.text import TextResponse
from scrapy.http.response.xml import XmlResponse
from scrapy.http.response import Response
from scrapy.selector.unified import SelectorList
from scrapy.item import Item
from crawlab.utils import OutputWidgetHandler

from ._version import version_info, __version__

from .widgets import *

crawler = Crawler()


def _jupyter_nbextension_paths():
    return [{
        'section': 'notebook',
        'src': 'static',
        'dest': 'crawlab',
        'require': 'crawlab/extension'
    }]


def configure_display():
    for resp_class in [HtmlResponse, TextResponse, XmlResponse]:
        resp_class._ipython_display_ = \
            lambda r: display(HTMLResponseWidget(r, crawler.get_messages(r.url), crawler.get_handler(r.url)))

    Response._ipython_display_ = \
        lambda r: display(ResponseWidget(r, crawler.get_messages(r.url), crawler.get_handler(r.url)))
    SelectorList._ipython_display_ = lambda s: display(SelectorWidget(s))
    Item._ipython_display_ = lambda i: display(JsonWidget(i))


def load_ipython_extension(ip):
    configure_display()
    mc = MagicCrawler(ip, crawler)
    ip.register_magics(mc)



import subprocess
import os
import html
from time import sleep
import logging

from zapv2 import ZAPv2

from crawlab.settings import ZAP_PATH

logger = logging.getLogger('Proxy')


class ZapProxy:
    def __init__(self, host='127.0.0.1', port='8080'):
        self.host = host
        self.port = port
        self.client = ZAPv2(apikey=None, proxies={'http': self.get_url()})
        self.started = False

    def get_url(self):
        return 'http://{}:{}'.format(self.host, self.port)

    def start_server(self):
        self.started = self.is_running()
        if not self.started:
            cmd = [ZAP_PATH, '-config api.disablekey=true', '-daemon', '-host', self.host, '-port', self.port]
            try:
                resp = subprocess.Popen(cmd, stdout=open(os.devnull, 'w'))
                self.started = self._wait_for_server()
                if not self.started:
                    logger.error('Proxy not started, process :' + str(resp))
            except FileNotFoundError:
                logger.error('ZAP proxy not found')
        return self.started

    def get_messages(self, site_url=None):
        messages = []
        if self.started:
            messages = self.client.core.messages(baseurl=site_url)
        return [self._format_message(message) for message in messages]

    def delete_messages(self, site_url):
        return self.client.core.delete_site_node(site_url)

    def is_running(self):
        try:
            self.client.urlopen(self.get_url())
        except Exception as e:
            logger.debug(e)
            return False
        return True

    def _format_message(self, message):
        return {
            'request': {
                'title': 'Request',
                'header': message['requestHeader'],
                'body': html.escape(message['requestBody'])
            },
            'response': {
                'title': 'Response',
                'header': message['responseHeader'],
                'body': html.escape(message['responseBody'])
            }
        }

    def _wait_for_server(self, max_seconds=20):
        logger.info('starting proxy server..')
        for i in range(max_seconds):
            if self.is_running():
                logger.info('started!')
                return True
            sleep(1)
        return False
